import importlib
import numbers
import socket
import queue
import threading
import asyncio
from typing import Union

import select
import functools

import dill
import numpy as np
import pandas as pd
from scipy.optimize import minimize as scipy_minimize
from scipy.optimize._minimize import MINIMIZE_METHODS as scipy_minimize_methods

from bayes_opt import BayesianOptimization, UtilityFunction

from empyric.tools import (
    convert_time,
    autobind_socket,
    read_from_socket,
    write_to_socket,
    get_ip_address,
    logger,
)
from empyric.types import (
    recast,
    Boolean,
    Integer,
    Float,
    Toggle,
    OFF,
    ON,
    Array,
    String,
)
from empyric.types import supported as supported_types


class Routine:
    """
    A Routine periodically updates a set of `knobs` based on a given state of an
    experiment. The knobs argument should be a dictionary of the form
    {..., name: variable, ...}.

    The optional `enable` argument should be a string corresponding to a key in whatever
    object is being passed as the `state` argument to the `update` method. The
    corresponding value should be boolean. When the enabling value is False, the
    `update` method takes no action. Otherwise, the `update` method proceeds normally.

    The optional `start`, `end` and `duration` arguments indicate when the routine
    should start and end. The `start` argument can be a number in seconds or a string
    with a number and units, e.g. "2 minutes", indicating when the routine starts, or it
    can be set to 'on enable' such that the start time is set to the time at which it is
    enabled (possibly repeatedly) and the end time is set to the start time plus
    `_duration` on each call to the `update` method. One can specify either `end` or
    `duration`, but not both; `end` is the absolute time at which the routine will end,
    while `duration` is the length of time the routine will run for (end minus start).

    All other arguments are fixed values or dictionary keys, corresponding to variable
    values of the controlling experiment.
    """

    assert_control = True
    #: if True, routine has sole control of knobs when active

    _start_on_enable = False

    _duration = 0.0

    def __init__(
        self,
        knobs: dict,
        enable: String = None,
        delay_prep: Boolean = False,
        start: Union[Float, String] = None,
        end: Union[Float, String] = None,
        duration: Union[Float, String] = None,
        **kwargs,
    ):
        self.knobs = knobs

        for knob in self.knobs.values():
            knob._controller = None  # to control access to knob

        self.enable = enable

        self.delay_prep = delay_prep

        if start is not None:
            if start == "on enable":
                self.start = np.nan  # will be set when routine is enabled
                self._start_on_enable = True
            else:
                self.start = convert_time(start)
        else:
            self.start = 0.0

        if end is not None:
            self.end = convert_time(end)
            self._duration = self.end - self.start
        elif duration is not None:
            self.end = self.start + convert_time(duration)
            self._duration = convert_time(duration)
        else:
            self.end = np.inf
            self._duration = np.inf

        self.prepped = False
        self.running = False
        self.finished = False

        for key, value in kwargs.items():
            self.__setattr__(key.replace(" ", "_"), value)

    @staticmethod
    def enabler(update):
        """
        Checks the enabling variable, start and stop times for the
        routine. If the enable variable evaluates to `True` and the time is
        between the start and stop times, the update method is called.

        If the enabling variable does not evaluate to `True`, then the routine's
        control of the knobs is revoked and no other action is taken.

        Otherwise, if the time is before the start time, the routine's prep
        method is called. If the time is after the end time, the routine's
        cleanup method is called.
        """

        @functools.wraps(update)
        def wrapped_update(self, state):
            if self.enable is not None and not state[self.enable]:
                # if routine has an enabling variable, and it evaluates to false...

                for name, knob in self.knobs.items():
                    # release control of knobs
                    if knob._controller == self:
                        knob._controller = None

                if self._start_on_enable:
                    # if start time is reset on enable, nullify start and end values
                    self.start = np.nan
                    self.end = np.nan

                self.running = False

                return

            elif state["Time"] < self.start:
                # prepare routine before starting, if needed
                if not self.prepped and not self.delay_prep:
                    self.prep(state)
                    self.prepped = True

                self.running = False

                return

            elif state["Time"] >= self.end:
                # finalize routine after ending, if needed
                if not self.finished:
                    self.finish(state)
                    self.finished = True

                    # Release knobs from control
                    for name, knob in self.knobs.items():
                        if knob._controller == self:
                            knob._controller = None

                self.running = False

                return

            else:
                # getting here means that the routine is running

                if not self.prepped:
                    self.prep(state)
                    self.prepped = True

                if self._start_on_enable and np.isnan(self.start):
                    # set start time to time of most recent state evaluation
                    self.start = state["Time"]
                    self.end = self.start + self._duration

                for name, knob in self.knobs.items():
                    if knob._controller and knob._controller != self:
                        # take no action if another routine has control
                        return
                    elif self.assert_control:
                        # assert control if needed
                        knob._controller = self

                self.running = True

                update(self, state)

        return wrapped_update

    @enabler
    def update(self, state):
        """
        Updates the knobs controlled by the routine based on the given state.

        Calls to this method are moderated by the `enabler` wrapper, which
        checks the start and end times as well as the enabling variable. The
        contents of this method just encode what happens when the routine is
        intended to be running.

        :param state: (dict/Series) state of the calling experiment or process
        in the form, {..., variable: value, ...}
        :return: None
        """

        pass

    def terminate(self):
        """
        Signals the routine to stop
        """

        pass

    def prep(self, state):
        """
        Does any needed preparation before the routine starts
        """

        pass

    def finish(self, state):
        """
        Makes any final actions after the routine ends
        """

        pass


class Set(Routine):
    """
    Sets `knobs` to the given values.

    The `values` argument should be either a single key/value for all knobs or
    a 1D array of keys/values of the same length as the `knobs` argument.
    """

    def __init__(self, knobs: dict, values, **kwargs):
        Routine.__init__(self, knobs, **kwargs)

        if len(self.knobs) > 1 and np.ndim(values) == 0:
            # One value for all knobs
            self.values = [values] * len(self.knobs)
        else:
            self.values = values

    @Routine.enabler
    def update(self, state):
        for knob, value in zip(self.knobs.values(), self.values):
            if isinstance(value, str):
                value = state[value]

            if value:
                knob.value = value


class Ramp(Routine):
    """
    Ramps the set of `knobs` to given `target` values at `given rates`.

    The `target` and `rate` arguments can be single values for all knobs or
    1D arrays of values. They can also be the names of variables, to be looked up from
    the `state` argument of the `update` method.
    """

    _start_on_enable = True

    def __init__(self, knobs: dict, targets, rates, **kwargs):
        """
        :param rates: (1D array) list of ramp rates
        """

        Routine.__init__(self, knobs, **kwargs)

        if np.ndim(rates) == 0:  # single rate for all knobs
            rates = [rates] * len(self.knobs)

        if np.ndim(targets) == 0:  # single target for all knobs
            targets = [targets] * len(self.knobs)

        self.rates = rates
        self.targets = targets

        self.now = None
        self.then = None  # last time of update call

    @Routine.enabler
    def update(self, state):
        self.now = state["Time"]

        if self.then is None or self.then < self.start:
            self.then = self.now
            return

        for knob, rate, target in zip(self.knobs, self.rates, self.targets):
            # target and rate can be variables
            if isinstance(target, String):
                target = state[target]

            if isinstance(rate, String):
                rate = state[rate]

            val_now = self.knobs[knob].value

            if (
                not isinstance(target, numbers.Number)
                or not isinstance(rate, numbers.Number)
                or not isinstance(val_now, numbers.Number)
                or target == val_now
            ):
                # Do nothing if knob, rate or target are undefined,
                # or if knob is already at target value
                continue

            sign = (target - val_now) / abs(target - val_now)

            val_nxt = val_now + sign * rate * (self.now - self.then)

            if sign * val_now <= sign * target < sign * val_nxt:
                self.knobs[knob].value = target
            else:
                self.knobs[knob].value = val_nxt

        self.then = self.now


class Timecourse(Routine):
    """
    Ramps the `knobs` linearly through a series of `values` at given `times`.

    The `times` and `values` should be 1D or 2D arrays of values (or keys for
    `values`); if 2D then the first dimension needs to have the same length as
    the `knobs` argument.
    """

    def __init__(self, knobs: dict, times, values, **kwargs):
        """

        :param times: (1D/2D array) array or list of times relative to
                      the start time
        :param values: (1D/2D array) array or list of values
        :param kwargs: keyword arguments for Routine
        """

        Routine.__init__(self, knobs, **kwargs)

        if np.ndim(times) == 0:  # single value or file path
            self.times = [[times]] * len(self.knobs)
        elif np.ndim(times) == 1:  # single list of times for all knobs
            self.times = [times] * len(self.knobs)
        else:
            self.times = times

        if np.ndim(values) == 0:  # single value or file path
            self.values = [[values]] * len(self.knobs)
        elif np.ndim(values) == 1:  # single list of values for all knobs
            self.values = [values] * len(self.knobs)
        else:
            self.values = values

        # Times and/or values can be stored in CSV files
        for i, (t_elem, v_elem) in enumerate(zip(self.times, self.values)):
            if type(t_elem[0]) == str and ".csv" in t_elem[0]:
                df = pd.read_csv(t_elem[0])

                try:
                    key = [col for col in df.columns if "time" in col.lower][0]
                except IndexError:
                    raise KeyError(f"No 'times' values found in {t_elem[0]}")

                self.times[i] = df[key].values

            if type(v_elem[0]) == str and ".csv" in v_elem[0]:
                df = pd.read_csv(v_elem[0])

                try:
                    key = [
                        col
                        for col in df.columns
                        if col == "values" or col == list(self.knobs)[i]
                    ][0]
                except IndexError:
                    raise KeyError(
                        f"No values for {list(self.knobs)[i]} found in {v_elem[0]}"
                    )

                self.values[i] = [recast(val) for val in df[key].values]

        self.times = np.array(convert_time(self.times)).astype(float)
        self.values = np.array(self.values, dtype=object)

        # Infer start and end times from times argument, if not given
        if "start" not in kwargs:
            self.start = np.min(self.times)

        if "end" not in kwargs:
            self.end = np.max(self.times)

        self.ramp = kwargs.get("ramp", True)

    @Routine.enabler
    def update(self, state):

        now = state["Time"]

        knobs_times_values = zip(self.knobs, self.times, self.values)

        for knob, times, values in knobs_times_values:
            if np.min(times) > now or np.max(times) < now:
                continue

            j_last = np.argwhere(times <= now).flatten()[-1]
            j_next = np.argwhere(times > now).flatten()[0]

            last_time = times[j_last]
            next_time = times[j_next]

            last_value = values[j_last]
            next_value = values[j_next]

            if last_value in list(state.keys()):
                # if last_value is a variable name,
                # use that variable's value from state
                last_value = state[last_value]

            if next_value in list(state.keys()):
                next_value = state[next_value]

            if self.ramp:
                # Ramp linearly between numerical values
                value = last_value + (next_value - last_value) * (now - last_time) / (
                    next_time - last_time
                )
            else:
                # or just set to last value
                value = last_value

            self.knobs[knob].value = value

    def prep(self, state):
        # Validate values
        for knob, value_list in zip(self.knobs.keys(), self.values):
            for value in value_list:
                is_number = isinstance(value, numbers.Number)
                is_variable = value in list(state.keys())

                if not is_number and not is_variable:
                    raise ValueError(
                        f"value {value} given for knob {knob} in Timecourse "
                        f"routine is invalid; value must be a numeric type or "
                        f"the name of a variable in the updating state"
                    )

    def finish(self, state):
        # Upon routine completion, set each knob to its final value
        for knob, value in zip(self.knobs.values(), self.values[:, -1]):
            if knob._controller is None or knob._controller == self:
                if isinstance(value, String) and value in state:
                    knob.value = state[value]
                else:
                    knob.value = value


class Sequence(Routine):
    """
    Passes the `knobs` through a series of `values` regardless of time.

    The `values` should be a 1D or 2D array of keys/values; if 2D then the first
    dimension needs to have the same length as the `knobs` argument.
    """

    def __init__(self, knobs: dict, values, **kwargs):
        Routine.__init__(self, knobs, **kwargs)

        if np.ndim(values) == 0:  # single value or file path
            self.values = [[values]] * len(self.knobs)
        elif np.ndim(values) == 1:  # single list of values for all knobs
            self.values = [values] * len(self.knobs)
        else:
            self.values = values

        # Values can be stored in CSV files
        for i, v_elem in enumerate(self.values):
            if type(v_elem[0]) == str and ".csv" in v_elem[0]:
                df = pd.read_csv(v_elem[0])

                try:
                    key = [
                        col
                        for col in df.columns
                        if col == "values" or col == list(self.knobs)[i]
                    ][0]
                except IndexError:
                    raise KeyError(
                        f"No values for {list(self.knobs)[i]} found in {v_elem[0]}"
                    )

                self.values[i] = [recast(val) for val in df[key].values]

        self.values = np.array(self.values, dtype=object)

        self.repeat = kwargs.get("repeat", True)

        self.iteration = 0

    @Routine.enabler
    def update(self, state):
        for knob, values in zip(self.knobs.values(), self.values):
            value = values[self.iteration]

            if isinstance(value, String):
                value = state[value]

            knob.value = value

        self.iteration = (self.iteration + 1) % len(self.values[0])

        if self.iteration == 0 and not self.repeat:
            self.end = state["Time"]  # set to end immediately

    def finish(self, state):
        # Upon routine completion, set each knob to its final value
        for knob, value in zip(self.knobs.values(), self.values[:, -1]):
            if knob._controller is None or knob._controller == self:
                if isinstance(value, String) and value in state:
                    knob.value = state[value]
                else:
                    knob.value = value


class Optimization(Routine):
    """
    Maximize or minimize a meter or expression influenced by the set of knobs.

    The `bounds` parameter is an Nx2 array containing the upper and lower limits for
    each of the N knobs that defines the parameter space over which the knob values can
    be explored.

    The `sign` argument determines whether the routine maximizes (sign = +1.0) or
    minimizes (-1.0) the `meter` value by tuning the `knobs`. The `Maximization` and
    `Minimization` subclasses have implied signs, so this argument should not be used
    with those.

    The `max_deltas` parameter is a value or 1-D array of values which are the
    maximum change in a single step the knob(s) can take.

    The `settling_time` is the amount of time to wait before changing knob settings and
    evaluating the meter.

    The `method` parameter specifies the algorith to use for optimization. The default
    method is a simple random walk search for the optimal knob settings. If `method` is
    set to 'bayesian' this routine uses a `bayesian optimizer
    <https://github.com/bayesian-optimization/BayesianOptimization>`_, which models
    the relation between the knobs and the meter as a gaussian process. Additionally,
    any of the methods available to the `scipy.optimize.maximize` function are also
    available; any keyword arguments accepted by that function can also be accepted by
    the constructor of the `Optimizer` class.

    The `iterations` parameter determines how many iterations the optimization
    algorithm is allowed before returning a result. When the iterations are complete,
    the algorithm is restarted anew.

    The `samples` parameter is how many values of the meter to average together on each
    iteration to evaluate the optimality of the current knobs settings.

    The default optimizer accepts a `recency` keyword argument, which is the weight
    given to the most recent best meter value at the best knob settings compared to
    previously obtained values. This can be useful when the meter is noisy and the
    relationship between the meter and knobs is changing over time.
    """

    _sign = None

    def __init__(
        self,
        knobs: dict,
        meter: String,
        bounds: Union[list, tuple, np.ndarray],
        sign: Float = None,
        max_deltas: Union[float, int, list, tuple, np.ndarray] = None,
        settling_time: Union[Float, String] = 0.0,
        method: String = None,
        iterations: int = 100,
        samples: int = 1,
        **kwargs,
    ):
        Routine.__init__(self, knobs, **kwargs)

        if self._sign is None:
            if sign == 1.0 or sign == -1.0:
                self._sign = sign
            else:
                raise ValueError(
                    "sign must be either +1.0 (maximization) or -1.0 (minimization)"
                )
        else:
            if sign is not None:
                logger.warning(
                    "The signs of Maximization and Minimization routines cannot be "
                    "changed"
                )

        self.bounds = {
            knob: subbounds
            for knob, subbounds in zip(knobs, np.reshape(bounds, (len(knobs), -1)))
        }

        if max_deltas:
            if np.ndim(max_deltas) == 0:
                # a single maximum change for all knobs
                self.max_deltas = np.array([max_deltas] * len(knobs))
            elif np.ndim(max_deltas) == 1 and len(max_deltas) == len(knobs):
                # individual maximum changes for each knob
                self.max_deltas = np.array(max_deltas)
            else:
                ValueError(
                    f"Improperly specified max_deltas parameter {max_deltas} for "
                    "optimization routine; must be either a single value or 1-D array "
                    "with the same length as the knobs argument"
                )
        else:
            self.max_deltas = np.array([np.inf] * len(self.knobs))

        # name of the meter to check from state argument of update
        self.meter = meter

        if "sign" in kwargs:
            self._sign = float(kwargs.pop("sign"))

        # Record best settings when an optimum is found
        self.best_meter = -self._sign * np.inf
        self.best_knobs = {name: None for name in self.knobs}

        self.method = method

        self.settling_time = convert_time(settling_time)

        self.iterations = iterations

        self.samples = int(samples)

        self._optimizer_thread = None
        self._meter_queue = queue.Queue(1)

        self.options = kwargs

        self._last_setting = -np.inf

        self._state = None

    @Routine.enabler
    def update(self, state):
        non_numeric_knobs = [
            not isinstance(state[knob], numbers.Number) for knob in self.knobs
        ]

        if np.any(non_numeric_knobs):
            # undefined state; take no action
            return

        if not isinstance(state[self.meter], numbers.Number):
            # undefined meter value; take no action
            return

        if state["Time"] < self._last_setting + self.settling_time:
            # still settling; take no action
            return

        # Reinitialize the optimizer, if iterations have completed
        if not self._optimizer_thread.is_alive():
            self.prep(state, start_optimizer_thread=True)

        self._state = state

        self._meter_queue.put(state[self.meter])

    def prep(self, state, start_optimizer_thread=False):
        if self.method == "bayesian":
            self._bayesian_optimizer = BayesianOptimization(
                f=lambda *args, **x: self._sign
                * self._eval_func([x[name] for name in self.knobs]),
                verbose=0,
                pbounds=self.bounds,
                random_state=6174,
                allow_duplicate_points=True,
            )

            self._bayesian_optimizer.probe(
                {name: knob.value for name, knob in self.knobs.items()}
            )

            optimizer_function = self._bayesian_optimizer.maximize
            optimizer_args = []
            optimizer_kwargs = {"init_points": 1, "n_iter": self.iterations}

        elif self.method in scipy_minimize_methods:
            options = {"maxiter": self.iterations}

            x0 = np.array([knob.value for knob in self.knobs.values()])

            if self.method.lower() == "nelder-mead":
                options["adaptive"] = True

                # Generate an initial simplex based on either max deltas or bounds
                bounds = np.array([(lb, ub) for lb, ub, in self.bounds.values()])

                if np.all(np.isfinite(self.max_deltas)):
                    d = self.max_deltas
                else:
                    d = 0.5 * np.diff(bounds, axis=1).flatten()

                N = len(x0)

                simplex = np.empty((N + 1, N), dtype=np.float64)
                simplex[0] = x0
                for k in range(N):
                    y = np.array(x0, copy=True)
                    if y[k] != 0:
                        y[k] = y[k] + d[k]
                    else:
                        y[k] = d[k]

                    y[k] = y[k] + d[k]

                    if y[k] < bounds[k][0]:
                        y[k] = bounds[k][0]
                    elif y[k] > bounds[k][1]:
                        y[k] = bounds[k][1]

                    simplex[k + 1] = y

                options["initial_simplex"] = simplex

            optimizer_function = scipy_minimize
            optimizer_args = [lambda x: -self._sign * self._eval_func(x), x0]
            optimizer_kwargs = {
                "method": self.method,
                "bounds": [bounds for bounds in self.bounds.values()],
                "options": options,
            }

        else:
            # use default optimization algorithm
            optimizer_function = self._default_optimization
            optimizer_args = [
                lambda x: self._eval_func(x),
                [knob.value for knob in self.knobs.values()],
            ]
            optimizer_kwargs = {
                "bounds": [bounds for bounds in self.bounds.values()],
                "max_deltas": self.max_deltas,
                "iterations": self.iterations,
                "sign": self._sign,
                "recency": self.options.get("recency", 1.0),
            }

        self._optimizer_thread = threading.Thread(
            target=optimizer_function, args=optimizer_args, kwargs=optimizer_kwargs
        )

        if start_optimizer_thread:
            self._optimizer_thread.start()
            self._optimizer_thread.started = True
        else:
            self._optimizer_thread.started = False

    def finish(self, state):
        # Shutdown _optimizer_thread
        if hasattr(self, "_meter_queue") and self._optimizer_thread.is_alive():
            self._meter_queue.put(StopIteration)

        # Apply optimal knob settings
        for i, (knob, value) in enumerate(self.best_knobs.items()):
            if value is None or not np.isfinite(value):
                print(f"Warning: No optimal value was found for {knob}")
            else:
                self.knobs[knob].value = value

    def terminate(self):
        # Shutdown _optimizer_thread
        if self._optimizer_thread is not None and self._optimizer_thread.is_alive():
            self._meter_queue.put(StopIteration)

    def _eval_func(self, x):
        """
        Emulates a function which takes knobs values as inputs and returns the
        resulting meter output, but only evaluates at most once per call to the
        update method.
        """

        for i, knob in enumerate(self.knobs.values()):
            knob.value = x[i]

        if self._state is not None:
            self._last_setting = self._state["Time"]

        values = []
        while len(values) < self.samples:
            # wait until invoked update method provides a new meter value
            new_value = self._meter_queue.get()

            if new_value == StopIteration:
                raise SystemExit  # terminate the enclosing _optimization_thread
            else:
                values.append(new_value)

        value = np.mean(values)

        # If this is the best configuration so far, store knob and meter values
        if self._sign * value > self._sign * self.best_meter:
            self.best_meter = value
            self.best_knobs = {name: knob.value for name, knob in self.knobs.items()}

        return value

    @staticmethod
    def _default_optimization(
        func, p0, bounds=None, max_deltas=None, sign=1.0, iterations=100, recency=1.0
    ):
        """Very basic random step optimization algorithm"""
        pbest = p = np.array(p0)
        fbest = func(p0)

        for i in range(iterations):
            # generate new knobs settings
            dp = np.array(
                [max_delta * (2 * np.random.rand() - 1) for max_delta in max_deltas]
            )

            p = p + dp

            if bounds:
                p = np.max([p, np.array(bounds)[:, 0]], axis=0)  # enforce lower bounds
                p = np.min([p, np.array(bounds)[:, 1]], axis=0)  # enforce upper bounds

            f = func(p)  # evaluate the meter
            if sign * f > sign * fbest:  # new best settings found; store values
                pbest = p
                fbest = f
            else:  # reevaluate meter at best knob settings
                p = pbest
                fbest = recency * func(p) + (1 - recency) * fbest

        return pbest, fbest


class Maximization(Optimization):
    """Alias for `Optimization` with `sign` implicitly set to +1.0"""

    _sign = +1.0


class Minimization(Optimization):
    """Alias for `Optimization` with `sign` implicitly set to -1.0"""

    _sign = -1.0


class SocketServer(Routine):
    """
    Server routine for transmitting data to other experiments, local or remote,
    using the socket interface.

    Any knobs given in the `knobs` argument can be read and set by clients.
    Clients can also read the values of any variables of the controlling
    experiment, via the `state` argument of the `update` method.
    """

    assert_control = False

    def __init__(self, knobs: dict = None, **kwargs):
        if knobs is None:
            knobs = {}

        Routine.__init__(self, knobs, **kwargs)

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self.ip_address, self.port = autobind_socket(self.socket)

        self.socket.listen(5)
        self.socket.settimeout(1)

        self.clients_queue = queue.Queue(1)
        self.clients_queue.put({})

        self.running = True

        self.state = {}

        self.server_thread = threading.Thread(
            target=asyncio.run, args=(self._run_async_server(),)
        )

        self.server_thread.start()

    def terminate(self):
        # Kill client handling threads
        self.running = False
        self.server_thread.join()

    async def _run_async_server(self):
        asyncio.create_task(self._accept_connections())
        asyncio.create_task(self._process_requests())

        async def server_forever():
            while self.running:
                await asyncio.sleep(0.1)

        await server_forever()

    async def _accept_connections(self):
        if self.running:
            try:
                (client, address) = self.socket.accept()

                client.settimeout(1)

                clients = self.clients_queue.get()

                clients[address] = client

                self.clients_queue.put(clients)

                print(f"Client at {address} has connected")

            except socket.timeout:
                pass

            await asyncio.sleep(0.001)

            asyncio.create_task(self._accept_connections())

        else:
            # Close sockets
            try:
                self.socket.shutdown(socket.SHUT_RDWR)
            except OSError:  # socket was not connected
                pass
            self.socket.close()

            clients = self.clients_queue.get()

            for address, client in clients.items():
                try:
                    client.shutdown(socket.SHUT_RDWR)
                    client.close()
                except ConnectionError:
                    # client is already disconnected
                    pass
                except OSError:
                    # client is already disconnected
                    pass

            # Return empty clients dict to queue so process_requests can exit
            self.clients_queue.put({})

    async def _process_requests(self):
        if self.running:
            clients = self.clients_queue.get()

            # Purge disconnected clients
            clients = {
                address: client
                for address, client in clients.items()
                if client is not None
            }

            for address, client in clients.items():
                outgoing_message = None

                try:
                    request = read_from_socket(client, chunk_size=1)
                except ConnectionError:
                    print(f"Connection issue with client at {address}")
                    client.shutdown(socket.SHUT_RDWR)
                    client.close()
                    clients[address] = None
                    continue

                if request:
                    alias = " ".join(request.split(" ")[:-1])
                    value = request.split(" ")[-1]

                    if alias not in self.knobs and alias not in self.state:
                        outgoing_message = f"Error: invalid alias"

                    elif value == "settable?":
                        settable = (alias in self.knobs) and self.knobs[alias].settable

                        if settable:
                            outgoing_message = f"{alias} settable"
                        elif alias in self.state:
                            outgoing_message = f"{alias} read-only"
                        else:
                            outgoing_message = f"{alias} undefined"

                    elif value == "type?":
                        if alias in self.knobs:
                            _type = self.knobs[alias]._type
                        elif alias in self.state:
                            _type = None
                            value = self.state[alias]

                            if isinstance(value, String) and ".csv" in value:
                                # value is an Array stored in a CSV file
                                _type = Array
                            else:
                                for supported_type in supported_types.values():
                                    if isinstance(value, supported_type):
                                        _type = supported_type

                        else:
                            _type = None

                        outgoing_message = f"{alias} {_type}"

                    elif value == "?":  # Query of value
                        if alias in self.knobs:
                            _value = self.knobs[alias].value
                        elif alias in self.state:
                            _value = self.state[alias]

                            if isinstance(_value, String) and ".csv" in _value:
                                # value is an array, list or tuple stored in CSV file
                                df = pd.read_csv(_value)

                                if alias in df.columns:
                                    # 1D array
                                    _value = df[alias].values
                                else:
                                    # 2D array
                                    _value = df.values

                        else:
                            _value = None

                        if isinstance(_value, Array):
                            # pickle arrays and send as bytes
                            outgoing_message = f"{alias} dlpkl".encode()
                            outgoing_message += dill.dumps(
                                _value, protocol=dill.HIGHEST_PROTOCOL
                            )
                        else:
                            outgoing_message = f"{alias} {_value}"

                    else:  # Setting a value
                        knob_exists = alias in self.knobs

                        is_free = not getattr(self.knobs[alias], "_controller", None)

                        if knob_exists and is_free:
                            knob = self.knobs[alias]

                            knob.value = recast(value, to=knob._type)

                            outgoing_message = f"{alias} {knob.value}"

                        else:
                            outgoing_message = f"Error: cannot set {alias}"

                # Send outgoing message
                if outgoing_message is not None:
                    write_to_socket(client, outgoing_message)
                    message_sent = True

                # Remove clients with problematic connections
                exceptional = client in select.select([], [], [client], 0)[2]

                if exceptional:
                    print(f"Client at {address} has a connection issue")
                    clients[address] = None

            self.clients_queue.put(clients)
            await asyncio.sleep(0.001)

            asyncio.create_task(self._process_requests())

    @Routine.enabler
    def update(self, state):
        self.state = state

    def __del__(self):
        if self.running:
            self.terminate()


class ModbusServer(Routine):
    """
    Server routine for transmitting data to other experiments, local or remote,
    using the Modbus over TCP/IP protocol.

    Any variables given in the `knobs` argument will be readable and writeable
    by connected clients; their values are stored in the holding registers. Variable
    values provided in the `state` argument of the `update` method can be read by
    clients; their values are stored in the input registers. The optional `meters`
    argument can be used to restrict the values from `state` that are readable by
    clients. Only the variable names listed in the `meters` argument will be accessible.
    If no `meters` argument is given, all variable values in the `state` argument will
    be readable.

    Because data is stored in statically assigned registers, only variables
    of boolean, toggle, integer or float types can be used. For more generic types, use
    the SocketServer routine.

    Values of writeable variables (the `knobs` argument) will be stored in
    holding registers in the same order as given in the argument, starting from
    address 0. Values provided in the `state` argument of the `update` method
    will be stored in input registers in the same order as defined therein, or in the
    same order of the `meters` argument if given, starting from address 0. Each value
    in both sets of registers is stored as 5 consecutive registers, 4 registers for the
    64-bit value and 1 register for metadata (i.e. data type). Note that the `state` of
    an instance of `Experiment` has "Time" as its first entry (i.e. input register 0 is
    the time value).
    """

    assert_control = False

    def __init__(
        self,
        knobs: dict = None,
        meters: Union[list, tuple, np.ndarray] = None,
        knob_addresses: Union[list, tuple, np.ndarray] = None,
        meter_addresses: Union[list, tuple, np.ndarray] = None,
        **kwargs,
    ):
        if knobs is None:
            knobs = {}

        Routine.__init__(self, knobs, **kwargs)

        if knob_addresses is not None:
            self.knob_addresses = knob_addresses
        else:
            self.knob_addresses = [5*i for i in range(len(knobs))]

        self.meters = meters

        if meters is not None:
            if meter_addresses is not None:
                self.meter_addresses = meter_addresses
            else:
                self.meter_addresses = [5*i for i in range(len(meters))]
        else:
            self.meter_addresses = []  # to be filled upon obtaining a state

        self.state = None  # set by update method

        # Check for PyModbus installation
        try:
            importlib.import_module("pymodbus")
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "pymodbus must be installed to run a Modbus server"
            )

        # Import tools from pymodbus
        datastore = importlib.import_module(".datastore", package="pymodbus")
        device = importlib.import_module(".device", package="pymodbus")
        payload = importlib.import_module(".payload", package="pymodbus")

        self._payload_builder = payload.BinaryPayloadBuilder
        self._payload_decoder = payload.BinaryPayloadDecoder

        DataBlock = datastore.ModbusSequentialDataBlock

        # Set up a PyModbus TCP Server

        self.slave = datastore.ModbusSlaveContext(
            ir=DataBlock.create(), hr=DataBlock.create()
        )

        self.slave.setValues = self.setValues_decorator(self.slave.setValues)

        self.context = datastore.ModbusServerContext(slaves=self.slave, single=True)

        self.identity = device.ModbusDeviceIdentification(
            info_name={
                "VendorName": "Empyric",
                "VendorUrl": "https://github.com/dmerthe/empyric",
                "ProductName": "Modbus Server",
                "ModelName": "Modbus Server",
            }
        )

        # Run server
        self.server = None  # assigned in _run_async_server
        self.ip_address = kwargs.get("address", get_ip_address())
        self.port = kwargs.get("port", 502)
        # TODO: warn when port is off-limits due to permissions
        self.server_thread = threading.Thread(
            target=asyncio.run, args=(self._run_async_server(),)
        )

        self.server_thread.start()

    def setValues_decorator(self, setValues_method):
        """
        This decorator adds a `from_vars` kwarg to indicate that the intention
        is to update the registers from variable values. Otherwise, the wrapped
        method sets the corresponding variables instead.
        """

        @functools.wraps(setValues_method)
        def wrapped_method(*args, from_vars=False):
            if from_vars:
                # update registers from variables
                return setValues_method(*args)
            else:
                # update variables according to the request
                fc_as_hex, address, values = args

                if fc_as_hex != 16:
                    logger.warning(
                        f"an attempt was made to write to a readonly "
                        f"register at address {address}"
                    )
                else:
                    if self.knobs:

                        # Identify knob corresponding to given register address
                        try:
                            index = [
                                i for i, addr in enumerate(self.knob_addresses)
                                if addr == address
                            ][0]
                        except IndexError:
                            logger.warning(
                                "an attempt was made to set a knob "
                                "through an unassigned register"
                            )
                            return

                        variable = list(self.knobs.values())[index]

                        controller = getattr(variable, "_controller", None)

                        if controller:
                            name = list(self.knobs.keys())[index]

                            logger.warning(
                                f"an attempt was made to set {name}, "
                                "but it is currently controlled by "
                                f"{controller}."
                            )
                            return

                        decoder = self._payload_decoder.fromRegisters(
                            values, byteorder=">"
                        )

                        if issubclass(variable._type, Boolean):
                            variable.value = decoder.decode_64bit_uint()
                        elif issubclass(variable._type, Toggle):
                            int_value = decoder.decode_64bit_uint()
                            variable.value = ON if int_value == 1 else OFF
                        elif issubclass(variable._type, Integer):
                            variable.value = decoder.decode_64bit_int()
                        elif issubclass(variable._type, Float):
                            variable.value = decoder.decode_64bit_float()

        return wrapped_method

    async def _update_registers(self):

        try:

            if self.state is None:  # do nothing if state is undefined
                await asyncio.sleep(0.1)
                return

            # Store readwrite variable values in holding registers (fc = 3)
            for i, (name, variable) in enumerate(self.knobs.items()):
                value = variable._value

                builder = self._payload_builder(byteorder=">")

                # encode the value into the 4 registers
                if value is None or variable._type is None:
                    builder.add_64bit_float(float("nan"))
                elif issubclass(variable._type, Boolean):
                    builder.add_64bit_uint(value)
                elif issubclass(variable._type, Toggle):
                    builder.add_64bit_uint(1 if value == ON else 0)
                elif issubclass(variable._type, Integer):
                    builder.add_64bit_int(value)
                elif issubclass(variable._type, Float):
                    builder.add_64bit_float(value)
                else:
                    raise ValueError(
                        f"unable to update modbus server registers from value "
                        f"{value} of variable {name} with data type "
                        f"{variable._type}"
                    )

                # encode the meta data
                meta_reg_val = {
                    Boolean: 0,
                    Toggle: 1,
                    Integer: 2,
                    Float: 3,
                    Array: 4,
                    String: 5,
                }.get(variable._type, -1)

                builder.add_16bit_int(meta_reg_val)

                # from_vars kwarg added with setValues_decorator above
                self.slave.setValues(
                    3, self.knob_addresses[i],
                    builder.to_registers(), from_vars=True
                )

            # Store readonly variable values in input registers (fc = 4)

            if self.meters is not None:

                try:
                    selection = {name: self.state[name] for name in self.meters}
                except KeyError as err:
                    raise ValueError(
                        f"meter {err.args[0]} not found in experiment variables"
                    )
            else:
                selection = self.state

            if len(self.meter_addresses) == 0:
                # assume consecutive sets of registers, if not specified
                self.meter_addresses = [5 * i for i in range(len(selection))]

            for i, (name, value) in enumerate(selection.items()):

                _type = None
                for supported_type in supported_types.values():
                    if isinstance(value, supported_type):
                        _type = supported_type

                builder = self._payload_builder(byteorder=">")

                # encode the value into the 4 registers
                if value is None or _type is None:
                    builder.add_64bit_float(float("nan"))
                elif _type == Boolean:
                    builder.add_64bit_uint(value)
                elif _type == Toggle:
                    builder.add_64bit_uint(int(value in Toggle.on_values))
                elif _type == Integer:
                    builder.add_64bit_int(value)
                elif _type == Float:
                    builder.add_64bit_float(value)
                else:
                    raise ValueError(
                        f"unable to update modbus server registers from value "
                        f"{value} of variable {name} with data type "
                        f"{_type}"
                    )

                # encode the meta data
                type_int = {Boolean: 0, Toggle: 1, Integer: 2, Float: 3}.get(_type, -1)

                builder.add_16bit_int(type_int)

                # from_vars kwarg added with setValues_decorator above
                self.slave.setValues(
                    4, self.meter_addresses[i],
                    builder.to_registers(), from_vars=True
                )

            await asyncio.sleep(0.1)

        finally:

            asyncio.create_task(self._update_registers())

    async def _run_async_server(self):

        asyncio.create_task(self._update_registers())

        server = importlib.import_module(".server", package="pymodbus")

        self.server = server.ModbusTcpServer(
            self.context, identity=self.identity, address=(self.ip_address, self.port)
        )

        try:
            await self.server.serve_forever()
        except asyncio.exceptions.CancelledError:
            # Server shutdown cancels the _update_registers task
            pass

    @Routine.enabler
    def update(self, state):
        self.state = state

    def terminate(self):

        # Close modbus server
        if not self.server.serving.done():
            self.server.serving.set_result(True)


supported = {
    key: value
    for key, value in vars().items()
    if type(value) is type and issubclass(value, Routine)
}



import time
import numpy as np

from empyric.types import Float
from empyric.adapters import Serial
from empyric.instruments import Instrument, measurer


class AlphaLabMR3(Instrument):
    """
    AlphaLab Inc. Magnetoresistive 3 Axis Milligauss Meter

    Accuracy of +/-0.5% of reading; dynamic range of +/-1999.9 milligauss
    """

    name = "AlphaLabMR3"

    supported_adapters = (
        (Serial, {"baud_rate": 115200, "timeout": 1.0, "delay": 1.0}),
    )

    meters = (
        "x component",  # x component of magnetic field
        "y component",
        "z component",
        "magnitude",  # magnitude of the magnetic field
    )

    def _measure_field(self) -> list:
        """Get the magnetic field vector"""

        # clear buffer
        while self.adapter.in_waiting:
            self.adapter.read(bytes=self.adapter.in_waiting, decode=False)
            time.sleep(0.1)

        response = self.query("\x03\x00\x00\x00\x00\x00", bytes=31, decode=False)

        _time = response[:6]
        vector = response[6:12], response[12:18], response[18:24]
        magnitude = response[24:30]
        end_byte = response[30]

        B_values = []

        for component in vector:

            sgn_dec_bits = format(component[1], "b")

            sgn_dec_bits = "0" * (8 - len(sgn_dec_bits)) + sgn_dec_bits

            sign = -1 if sgn_dec_bits[4] == "1" else 1

            decimal_places = sum(
                [int(bit) * 2 ** (2 - i) for i, bit in enumerate(sgn_dec_bits[-3:])]
            )

            int_value = int.from_bytes(component[-2:], byteorder="big")

            numerical_value = sign * int_value * 10 ** (-decimal_places)

            B_values.append(numerical_value)

        return B_values

    @measurer
    def measure_x_component(self) -> Float:
        """Measure the x component of the magnetic field"""
        return self._measure_field()[0]

    @measurer
    def measure_y_component(self) -> Float:
        """Measure the y component of the magnetic field"""
        return self._measure_field()[1]

    @measurer
    def measure_z_component(self) -> Float:
        """Measure the z component of the magnetic field"""
        return self._measure_field()[2]

    @measurer
    def measure_magnitude(self) -> Float:
        """Measure the magnetic field vector and compute the magnitude"""
        return np.linalg.norm(self._measure_field())



import re
import struct
import time
import numpy as np

from empyric.tools import find_nearest
from empyric.types import Float, Array, Integer, String, Boolean
from empyric.adapters import USB, Socket
from empyric.collection.instrument import Instrument, setter, getter, measurer


class TekTDSScope(Instrument):
    """
    Tektronix oscillscope of the TDS200, TDS1000/2000, TDS1000B/2000B,
    TPS2000 series.

    Although there are up to four channels, the constructor checks the model
    number for the actual number of supported channels. With the 2-channel
    models, any methods relating to channels 3 or 4 will take no action.

    2-Channel models: TXX1001, TXX1002, TXX1012, TXX2002, TXX2012, TXX2022

    4-channel models: TXX2004, TXX2014
    """

    name = "TekScope"

    supported_adapters = ((USB, {"timeout": 10}),)

    knobs = (
        "horz scale",
        "horz position",
        "ch1 scale",
        "ch1 position",
        "ch2 scale",
        "ch2 position",
        "ch3 scale",
        "ch3 position",
        "ch4 scale",
        "ch4 position",
        "trigger level",
    )

    meters = (
        "channel 1",
        "channel 2",
        "channel 3",
        "channel 4",
    )

    channels = 0

    two_channel_models = [1001, 1002, 1012, 2002, 2012, 2022]
    four_channel_models = [2004, 2014]

    def __init__(self, *args, **kwargs):
        # Check for number of channels before standard initialization
        adapter = USB(Instrument(args[0]))

        self.model = int(re.search("\d\d\d\d", adapter.query("*IDN?"))[0])

        adapter.disconnect()

        if self.model in self.two_channel_models:
            self.channels = 2
        elif self.model in self.four_channel_models:
            self.channels = 4

        super().__init__(*args, **kwargs)

    # Horizontal

    @setter
    def set_horz_scale(self, scale: Float):
        self.write("HOR:SCA %.3e" % scale)

    @getter
    def get_horz_scale(self) -> Float:
        return float(self.query("HOR:SCA?"))

    @setter
    def set_horz_position(self, position: Float):
        self.write("HOR:POS %.3e" % position)

    @getter
    def get_horz_position(self) -> Float:
        return float(self.query("HOR:POS?"))

    # Channel 1

    @setter
    def set_ch1_scale(self, scale: Float):
        self.write("CH1:SCA %.3e" % scale)

    @getter
    def get_ch1_scale(self) -> Float:
        return float(self.query("CH1:SCA?"))

    @setter
    def set_ch1_position(self, position: Float):
        self.write("CH1:POS %.3e" % position)

    @getter
    def get_ch1_position(self) -> Float:
        return float(self.query("CH1:POS?"))

    @measurer
    def measure_channel_1(self) -> Array:
        return self._measure_channel(1)

    # Channel 2

    @setter
    def set_ch2_scale(self, scale: Float):
        self.write("CH2:SCA %.3e" % scale)

    @getter
    def get_ch2_scale(self) -> Float:
        return float(self.query("CH2:SCA?"))

    @setter
    def set_ch2_position(self, position: Float):
        self.write("CH2:POS %.3e" % position)

    @getter
    def get_ch2_position(self) -> Float:
        return float(self.query("CH2:POS?"))

    @measurer
    def measure_channel_2(self) -> Array:
        return self._measure_channel(2)

    # Channel 3

    @setter
    def set_ch3_scale(self, scale: Float):
        if self.channels > 2:
            self.write("CH3:SCA %.3e" % scale)
        else:
            return np.nan

    @getter
    def get_ch3_scale(self) -> Float:
        if self.channels > 2:
            return float(self.query("CH3:SCA?"))
        else:
            return np.nan

    @setter
    def set_ch3_position(self, position: Float):
        if self.channels > 2:
            self.write("CH3:POS %.3e" % position)
        else:
            return np.nan

    @getter
    def get_ch3_position(self) -> Float:
        if self.channels > 2:
            return float(self.query("CH3:POS?"))
        else:
            return np.nan

    @measurer
    def measure_channel_3(self) -> Array:
        if self.channels > 2:
            return self._measure_channel(3)
        else:
            return [np.nan]

    # Channel 4

    @setter
    def set_ch4_scale(self, scale: Float):
        if self.channels > 2:
            self.write("CH4:SCA %.3e" % scale)
        else:
            return np.nan

    @getter
    def get_ch4_scale(self) -> Float:
        if self.channels > 2:
            return float(self.query("CH4:SCA?"))
        else:
            return np.nan

    @setter
    def set_ch4_position(self, position: Float):
        if self.channels > 2:
            self.write("CH4:POS %.3e" % position)
        else:
            return np.nan

    @getter
    def get_ch4_position(self) -> Float:
        if self.channels > 2:
            return float(self.query("CH4:POS?"))
        else:
            return np.nan

    @measurer
    def measure_channel_4(self) -> Array:
        if self.channels > 2:
            return self._measure_channel(4)
        else:
            return [np.nan]

    # Trigger

    @setter
    def set_trigger_level(self, level: Float):
        self.write("TRIG:MAI:LEV %.3e" % level)

    @getter
    def get_trigger_level(self) -> Float:
        return float(self.query("TRIG:MAI:LEV?"))

    def _measure_channel(self, n):
        self.write("DAT:ENC ASCI")  # ensure ASCII encoding of data
        self.write("DAT:SOU CH%d" % n)  # switch to channel n

        scale_factor = float(self.query("WFMPRE:YMULT?"))
        zero = float(self.query("WFMPRE:YZERO?"))
        offset = float(self.query("WFMPRE:YOFF?"))

        self.write("ACQ:STATE RUN")  # acquire the waveform

        while int(self.query("BUSY?")):
            time.sleep(0.25)  # wait for acquisition to complete

        normal_timeout = self.adapter.timeout
        self.adapter.timeout = 60

        response = self.query("CURVE?")

        str_data = response.split(",")

        self.adapter.timeout = normal_timeout

        return np.array(
            [(float(datum) - offset) * scale_factor + zero for datum in str_data]
        )


class MulticompProScope(Instrument):
    """
    Multicomp Pro PC Oscilloscope.
    """

    supported_adapters = (
        (Socket, {'timeout': 1.0}),
        (USB, {"delay": 1}),
        # acquisitions can take a long time
    )
    """Supported adapters and options."""

    knobs = (
        "horz scale",
        "horz position",
        "scale ch1",
        "position ch1",
        "scale ch2",
        "position ch2",
        "sweep mode",
        "trigger level",
        "trigger source",
        "sweep mode",
        "resolution",
        "acquire",
        "state",
    )

    meters = (
        "channel 1",
        "channel 2",
    )

    presets = {
        "resolution": 1e4,
        "sweep mode": "SINGLE",
        "trigger source": 1,
    }

    _time_scales = {
        2e-9: "2.0ns",
        5e-9: "5.0ns",
        10e-9: "10ns",
        20e-9: "20ns",
        50e-9: "50ns",
        100e-9: "100ns",
        200e-9: "200ns",
        500e-9: "500ns",
        1e-6: "1.0us",
        2e-6: "2.0us",
        5e-6: "5.0us",
        10e-6: "10us",
        20e-6: "20us",
        50e-6: "50us",
        100e-6: "100us",
        200e-6: "200us",
        500e-6: "500us",
        1e-3: "1.0ms",
        2e-3: "2.0ms",
        5e-3: "5.0ms",
        10e-3: "10ms",
        20e-3: "20ms",
        50e-3: "50ms",
        100e-3: "100ms",
        200e-3: "200ms",
        500e-3: "500ms",
        1: "1.0s",
        2: "2.0s",
        5: "5.0s",
        10: "10s",
        20: "20s",
        50: "50s",
        100: "100s",
    }

    _volt_scales = {
        2e-3: "2mv",
        5e-3: "5mv",
        10e-3: "10mv",
        20e-3: "20mv",
        50e-3: "50mv",
        100e-3: "100mv",
        200e-3: "200mv",
        500e-3: "500mv",
        1.0: "1v",
        2.0: "2v",
        5.0: "5v",
    }

    _volt_div_table = {0: 1e-3, 9: 1.0}

    _resolutions = {1e3: "1K", 1e4: "10K", 1e5: "100K", 1e6: "1M", 1e7: "10M"}

    _sweep_modes = ["AUTO", "NORMAL", "NORM", "SINGLE", "SING"]

    _acquiring = False

    sample_rate = None

    @setter
    def set_horz_scale(self, scale: Float):
        new_time_scale = find_nearest(list(self._time_scales.keys()), scale)

        self.write(":HORI:SCAL " + self._time_scales[new_time_scale])

    @getter
    def get_horz_scale(self) -> Float:
        time_scales_rev = {val: key for key, val in self._time_scales.items()}

        time_scale_str = self.query(":HORI:SCAL?")[:-2]

        return time_scales_rev[time_scale_str]

    @setter
    def set_horz_position(self, offset: Float):
        self.write(":HORI:OFFS " + str(offset))

    @getter
    def get_horz_position(self) -> Float:
        return float(self.query(":HORI:OFFS?")[:-2])

    def _set_scale_ch(self, channel, scale):
        new_volt_scale = find_nearest(list(self._volt_scales.keys()), scale)

        scale_str = self._volt_scales[new_volt_scale]

        self.write((":CH%d:SCAL " % channel) + scale_str)

    def _get_scale_ch(self, channel):
        volt_scale_str = self.query(":CH%d:SCAL?" % channel)[:-2]

        if "mv" in volt_scale_str.lower():
            return 1e-3 * float(volt_scale_str[:-2])
        elif "v" in volt_scale_str.lower():
            return float(volt_scale_str[:-1])
        else:
            return float("nan")

    def _set_position_ch(self, channel, position):
        self.write((":CH%d:OFFS " % channel) + str(position))

    @setter
    def set_scale_ch(self, scale: Float, channel=None):
        self._set_scale_ch(channel, scale)

    @getter
    def get_scale_ch(self, channel=None) -> Float:
        return self._get_scale_ch(channel)

    @setter
    def set_position_ch(self, position: Float, channel=None):
        self._set_position_ch(channel, position)

    @getter
    def get_position_ch(self, channel=None) -> Float:
        return float(self.query(":CH%d:OFFS?" % channel)[:-2])

    @setter
    def set_sweep_mode(self, mode: String):
        if mode.upper() not in self._sweep_modes:
            raise ValueError(
                f"Invalid sweep mode! "
                f"Sweep mode must be one of: {', '.join(self._sweep_modes)}"
            )

        self.write(":TRIG:SING:SWE " + mode.upper())

    @getter
    def get_sweep_mode(self) -> String:
        return self.query(":TRIG:SING:SWE?")[:-2].upper()

    @setter
    def set_trigger_source(self, source: String):
        self.write(":TRIG:SING:EDGE:SOUR CH%d" % int(source))

    @getter
    def get_trigger_source(self) -> String:
        source = self.query(":TRIG:SING:EDGE:SOUR?")[:-2]

        if source == "CH1":
            return 1
        elif source == "CH2":
            return 2

    @setter
    def set_trigger_level(self, level_volts: Float):
        trigger_source = self.get_trigger_source()

        scale = self._get_scale_ch(trigger_source)

        level_div = str(level_volts / scale)

        self.write(":TRIG:SING:EDGE:LEV " + level_div)

    @getter
    def get_trigger_level(self) -> Float:
        trigger_source = self.get_trigger_source()

        scale = self._get_scale_ch(trigger_source)

        return float(self.query(":TRIG:SING:EDGE:LEV?")[:-2]) * scale

    @measurer
    def measure_trigger_status(self) -> String:
        return self.query(":TRIG:STATUS?")[:-2]

    @setter
    def set_resolution(self, resolution: Integer):
        if resolution not in self._resolutions:
            raise ValueError(
                f"Invalid memory depth! "
                f"Memory depth must be one of: {', '.join(str(self._resolutions))}"
            )

        self.write(":ACQ:DEPMEM " + self._resolutions[resolution])

    @getter
    def get_resolution(self) -> Integer:
        res_rev = {val: key for key, val in self._resolutions.items()}

        return res_rev[self.query(":ACQ:DEPMEM?")[:-2]]

    @setter
    def set_state(self, state: String):
        if state.upper() in ["RUN", "STOP"]:
            self.write(f":{state.upper()}")

    @setter
    def set_acquire(self, _: Boolean):
        self._acquiring = True

        self.set_sweep_mode("SINGLE")

        trig_status = self.measure_trigger_status()

        if trig_status == "TRIG":
            # measurement has been triggered
            pass
        else:
            # sweep needs to be reset
            time.sleep(0.1)
            self.set_acquire(1)

    def _read_preamble(self, channel):
        info_dict = {}

        preamble = self.query(":WAV:PRE?", decode=False)

        header, info = preamble[:11], preamble[11:]

        if header != b"#9000000000":  # indicates empty data buffer
            try:
                # First 8 bytes is an integer signature, used for verification
                ver_int = 651058244139746640
                verified = struct.unpack("<q", info[:8])[0] == ver_int

                if verified:
                    if channel == 1:
                        scale = self._volt_div_table[
                            struct.unpack("<h", info[260:262])[0]
                        ]
                        zero = struct.unpack("<f", info[268:272])[0]
                    else:
                        scale = self._volt_div_table[
                            struct.unpack("<h", info[262:264])[0]
                        ]
                        zero = struct.unpack("<f", info[272:276])[0]

                    sample_rate = struct.unpack("<f", info[316:320])[0] * 1e6
                    # Hz

                    info_dict["scale"] = scale
                    info_dict["zero"] = zero
                    info_dict["sample rate"] = sample_rate
                else:
                    print(
                        f"Warning: {self.name} received unverified "
                        f"signature when reading from channel {channel}"
                    )

            except struct.error:
                print(
                    f"Warning: {self.name} received truncated preamble when "
                    f"reading from channel {channel}"
                )

        self.sample_rate = info_dict["sample rate"]

        return info_dict

    def _read_data(self, channel):
        info = self._read_preamble(channel)

        if info:
            scale = info["scale"]
            zero = info["zero"]
        else:
            return None

        # Only 256k data points can be read at a time
        resolution = self.get_resolution()

        if resolution < 256000:  # can get all data in one pass
            self.write(":WAV:RANG 0, %d" % resolution)

            raw_response = self.query(":WAV:FETC?", decode=False)

            header, byte_data = raw_response[:11], raw_response[11:]

            data_len = int(header[2:]) // 2

            voltages = scale * (
                np.array(struct.unpack(f"<{data_len}h", byte_data), dtype=float) / 6400
                - zero
            )

        else:  # need to read data in chunks; max chunk size is 256k
            offset, size = 0, 200000

            voltages = np.empty(resolution)

            while offset + size <= resolution:
                self.write(":WAV:RANG %d, %d" % (offset, size))

                raw_response = self.query(":WAV:FETC?", decode=False)

                header, byte_data = raw_response[:11], raw_response[11:]

                data_len = int(header[2:]) // 2

                voltages[offset: offset + size] = scale * (
                    np.array(struct.unpack(f"<{data_len}h", byte_data), dtype=float)
                    / 6400
                    - zero
                )

                offset += size

        if len(voltages) != resolution:
            print(
                f"Warning: {self.name} received truncated data "
                f"for channel {channel}; discarding"
            )
            return None

        return voltages

    @measurer
    def measure_channel_1(self) -> Array:
        if not self._acquiring:
            self.set_acquire(True)

        self._acquiring = False

        self.write(":WAV:BEG CH1")

        data = self._read_data(1)

        self.write(":WAV:END CH1")

        return data

    @measurer
    def measure_channel_2(self) -> Array:
        if not self._acquiring:
            self.set_acquire(True)

        self._acquiring = False

        self.write(":WAV:BEG CH2")

        data = self._read_data(2)

        self.write(":WAV:END CH2")

        return data


class SiglentSDS1000(Instrument):
    """
    Siglent SDS1000 series digital oscilloscope
    """

    name = "SiglentSDS1000"

    supported_adapters = (
        (
            Socket,
            {
                "write_termination": "\n",
                "read_termination": "\n",
                "timeout": 60,  # waveform retrieval may take some extra time
            },
        ),
    )

    knobs = (
        "horz scale",
        "horz position",
        "ch1 scale",
        "ch1 position",
        "ch2 scale",
        "ch2 position",
        "ch3 scale",
        "ch3 position",
        "ch4 scale",
        "ch4 position",
        "trigger source",
        "trigger level",
        "trigger mode",
        "arm",
        "acquire mode",
        "memory size",
        "averages",
    )

    meters = ("ch1 waveform", "ch2 waveform", "ch3 waveform", "ch4 waveform")

    horz_scales = [
        1e-9,
        2e-9,
        5e-9,
        10e-9,
        20e-9,
        50e-9,
        100e-9,
        200e-9,
        500e-9,
        1e-6,
        2e-6,
        5e-6,
        10e-6,
        20e-6,
        50e-6,
        100e-6,
        200e-6,
        500e-6,
        1e-3,
        2e-3,
        5e-3,
        10e-3,
        20e-3,
        50e-3,
        100e-3,
        200e-3,
        500e-3,
        1.0,
        2.0,
        5.0,
        10.0,
        20.0,
        50.0,
        100.0,  # seconds
    ]

    _trg_src = None

    # Time axis control
    @setter
    def set_horz_scale(self, scale: Float):
        nearest_scale = find_nearest(self.horz_scales, scale, overestimate=True)

        self.write(f"TDIV {nearest_scale}S")

        return nearest_scale

    @getter
    def get_horz_scale(self) -> Float:
        response = self.query("TDIV?").split("TDIV ")[-1][:-1]

        try:
            response = float(response)
        except ValueError:
            return np.nan

        return response

    @setter
    def set_horz_position(self, position: Float):
        self.write(f"TRDL {position}S")

    @getter
    def get_horz_position(self) -> Float:
        response = self.query("TRDL?").split("TRDL ")[-1][:-1]

        try:
            response = float(response)
        except ValueError:
            return np.nan

        return response

    # Channel control base functions
    def _set_chn_scale(self, n, scale):
        self.write("C%d:VDIV %.3eV" % (n, float(scale)))

    def _get_chn_scale(self, n):
        response = self.query("C%d:VDIV?" % n).split("C%d:VDIV " % n)[-1][:-1]

        try:
            response = float(response)
        except ValueError:
            return np.nan

        return response

    def _set_chn_position(self, n, position):
        self.write("C%d:OFST %.3eV" % (n, float(position)))

    def _get_chn_position(self, n):
        response = self.query("C%d:OFST?" % n).split("C%d:OFST " % n)[-1][:-1]

        try:
            response = float(response)
        except ValueError:
            return np.nan

        return response

    # Channel 1 control
    @setter
    def set_ch1_scale(self, scale: Float):
        self._set_chn_scale(1, scale)

    @getter
    def get_ch1_scale(self) -> Float:
        return self._get_chn_scale(1)

    @setter
    def set_ch1_position(self, position: Float):
        self._set_chn_position(1, position)

    @getter
    def get_ch1_position(self) -> Float:
        return self._get_chn_position(1)

    # Channel 2 control
    @setter
    def set_ch2_scale(self, scale: Float):
        self._set_chn_scale(2, scale)

    @getter
    def get_ch2_scale(self) -> Float:
        return self._get_chn_scale(2)

    @setter
    def set_ch2_position(self, position: Float):
        self._set_chn_position(2, position)

    @getter
    def get_ch2_position(self) -> Float:
        return self._get_chn_position(2)

    # Channel 3 control
    @setter
    def set_ch3_scale(self, scale: Float):
        self._set_chn_scale(3, scale)

    @getter
    def get_ch3_scale(self) -> Float:
        return self._get_chn_scale(3)

    @setter
    def set_ch3_position(self, position: Float):
        self._set_chn_position(3, position)

    @getter
    def get_ch3_position(self) -> Float:
        return self._get_chn_position(3)

    # Channel 4 control
    @setter
    def set_ch4_scale(self, scale: Float):
        self._set_chn_scale(4, scale)

    @getter
    def get_ch4_scale(self) -> Float:
        return self._get_chn_scale(4)

    @setter
    def set_ch4_position(self, position: Float):
        self._set_chn_position(4, position)

    @getter
    def get_ch4_position(self) -> Float:
        return self._get_chn_position(4)

    # Trigger control
    @setter
    def set_trigger_level(self, level: Float):
        trg_src = self.get_trigger_source()

        self.write(f"C{trg_src}:TRLV {level}V")

    @getter
    def get_trigger_level(self) -> Float:
        trg_src = self.get_trigger_source()

        if trg_src > 0:
            response = self.query(f"C{trg_src}:TRLV?").split("TRLV ")[-1][:-1]
        else:
            return np.nan

        try:
            response = float(response)
        except ValueError:
            return np.nan

        return response

    @setter
    def set_trigger_source(self, source: Integer):
        self.write(f"TRSE EDGE,SR,C{int(source)},OFF")

    @getter
    def get_trigger_source(self) -> Integer:
        def validator(response):
            return "SR,C" in response

        response = self.query("TRSE?", validator=validator).split("SR,C")[-1][0]

        try:
            response = int(response)
        except ValueError:
            return -1

        return response

    @setter
    def set_trigger_mode(self, mode: String):
        if mode.upper() in ["AUTO", "NORM", "SINGLE", "STOP"]:
            self.write(f"TRMD {mode.upper()}")
        else:
            return self.get_trigger_mode()

    @getter
    def get_trigger_mode(self) -> String:
        try:
            return self.query("TRMD?")[5:]
        except IndexError:
            return None

    @setter
    def set_acquire_mode(self, mode: String):
        if mode.upper() in ["SAMPLING", "PEAK_DETECT", "HIGH_RES"]:
            self.write(f"ACQW {mode.upper()}")
        elif mode.upper() == "AVERAGE":
            if not hasattr(self, "averages"):
                self.set_averages(4)

            self.write(f"ACQW AVERAGE,{int(self.averages)}")

        else:
            return None

    @getter
    def get_acquire_mode(self) -> String:
        response = self.query("ACQW?").split("ACQW ")[-1]

        if "AVERAGE" in response:  # also includes # of averages
            return "AVERAGE"
        else:
            return response

    @setter
    def set_memory_size(self, size: Integer):
        self.write(f"MSIZ {int(size)}")

        return self.get_memory_size()

    @getter
    def get_memory_size(self) -> Integer:
        response = self.query("MSIZ?").split("MSIZ ")[-1]

        if "K" in response:
            response = float(response.replace("K", "e3"))
        elif "M" in response:
            response = float(response.replace("M", "e6"))

        try:
            response = int(response)
        except ValueError:
            return -1

        return response

    @setter
    def set_averages(self, averages: Integer):
        valid = [4] + [2**i for i in range(4, 11)]

        averages = find_nearest(valid, averages)

        self.write(f"AVGA {averages}")

    @getter
    def get_averages(self) -> Integer:
        response = self.query("AVGA?").split("AVGA ")[-1]

        try:
            response = int(response)
        except ValueError:
            return -1

        return response

    # Channel measurements
    def _measure_chn_waveform(self, n):
        ready = False
        while not ready:
            time.sleep(0.25)
            ready = "1" in self.query("*OPC?")

        # Enable channel if needed
        ch_enabled_query = self.query("C%d:TRA?" % n)
        if "ON" not in ch_enabled_query:
            print(f"Warning: CH{n} is not enabled; cannot retrieve waveform")
            return None

        self.write("WFSU SP,0,NP,0,FP,0")  # setup to get all data points

        def is_terminated(message):
            try:
                size = int(message[13:22])
            except ValueError:
                return False

            data_length = len(message[22:-2])

            if data_length == size:
                return True
            else:
                return False

        self.adapter.read_termination = is_terminated
        max_attempts = self.adapter.max_attempts
        self.adapter.max_attempts = 1

        def validator(response):
            """
            Checks for correct header, size, message length and termination.
            """

            if response == b"":
                return True  # no data

            if response is None:
                return False
            elif len(response) == 0:
                return False

            header = response[:13]
            if header != (b"C%d:WF DAT2,#9" % n):
                print("invalid header", header)
                return False

            size = response[13:22]
            try:
                size = int(size)
            except ValueError:
                print("invalid data size", size)
                return False

            data = response[22:-2]
            if len(data) != size:
                print("truncated data", len(data))
                return False

            termination = response[-2:]
            if termination != b"\n\n":
                print("no termination", response[-2:])
                return False

            return True

        response = self.query("C%d:WF? DAT2" % n, decode=False, validator=validator)

        if response == b"":
            return None

        self.adapter.read_termination = "\n"  # restore normal read termination
        self.adapter.max_attempts = max_attempts

        header, size, waveform = response[:13], response[13:22], response[22:-2]
        size = int(size)

        # convert bytes to integers
        waveform = np.array(struct.unpack("b" * size, waveform), dtype=np.float64)

        scale = self._get_chn_scale(n)
        pos = self._get_chn_position(n)

        waveform = (scale / 25.0) * waveform - pos

        return waveform

    @measurer
    def measure_ch1_waveform(self) -> Array:
        return self._measure_chn_waveform(1)

    @measurer
    def measure_ch2_waveform(self) -> Array:
        return self._measure_chn_waveform(2)

    @measurer
    def measure_ch3_waveform(self) -> Array:
        return self._measure_chn_waveform(3)

    @measurer
    def measure_ch4_waveform(self) -> Array:
        return self._measure_chn_waveform(4)

    @measurer
    def measure_sample_rate(self) -> Float:
        response = self.query("SARA?")[5:-4]

        try:
            response = float(response)
        except ValueError:
            return np.nan

        return response
